{-# LANGUAGE RecordWildCards #-}

module Pipes.IRC.Server.EventHandler
  ( ircEventHandler )
where

import           Control.Applicative      ((<$>))
import           Control.Concurrent.Async
import           Control.Concurrent.STM
import           Control.Lens
import           Control.Monad
import           Data.Map                 as M
import           Data.Maybe               as DM
import           Pipes.Concurrent         as PC
import           Pipes.IRC.Server.Log
import           Pipes.IRC.Server.Types

sendToMany :: a -> [Output a] -> IO ()
sendToMany msg outs = do
  resL <- forM outs $ \o ->
    async $ atomically $ PC.send o msg
  mapM_ wait resL

ircEventHandler :: ServerState -> IrcEvent -> IO Bool
ircEventHandler srv evt =
  case evt of
    Close connId -> do
      outConns <- readTVarIO $ srv ^. ircConnections
      return $ isJust (M.lookup connId outConns)
    Msg {..} -> do
      outConns <- readTVarIO $ srv ^. ircConnections
      let os = _out <$> DM.mapMaybe (`M.lookup` outConns) _outDest
      sendToMany _outMsg os
      logOutMsg _outMsg _outDest
      return True
    Pong {..} -> do
      atomically $ modifyTVar' (srv ^. ircConnections) $
                     M.adjust (gotPong .~ True) _pongConn
      return True
